-- by modelleicher
-- www.ls-modcompany.com
-- 03.01.2020 

-- maizePlus: A global mod/script that adds the possibility to differentiate between maize, grass and wcs silage. Also adds the ability to bale maize silage.
-- also adds CCM 

-- Big thanks to Wopster for his help with making this script work without conflict due to heap indexing issues and with other stuff. Also I took a bunch of inspiration from seasons.

MaizePlus = {}

local modDirectory = g_currentModDirectory
local modName = g_currentModName

MaizePlus.isActive = true
MaizePlus.ccmIsEnabled = false

function MaizePlus.fillTypeInit(directory, fillTypeManager, i18n)

	-- enable CCM if CCM Extension is active 
	if g_modIsLoaded.FS19_MaizePlus_ccmExtension then
		MaizePlus.ccmIsEnabled = true
	end
	
    MaizePlus.loadChoppedMaizeFillType(directory, fillTypeManager, i18n)
    MaizePlus.loadChoppedMaizeFermentedFillType(directory, fillTypeManager, i18n)
    MaizePlus.loadGrassFermentedFillType(directory, fillTypeManager, i18n)
	MaizePlus.loadSiloGrass(directory, fillTypeManager, i18n)	

	if MaizePlus.ccmIsEnabled then
		MaizePlus.loadCCMFillType(directory, fillTypeManager, i18n)
		MaizePlus.loadCCMRawFillType(directory, fillTypeManager, i18n)
	end
end

function MaizePlus.loadChoppedMaizeFillType(directory, fillTypeManager, i18n)
    local hudFile = "hud/hud_fill_choppedMaize.png"
    local hudFileSmall = "hud/hud_fill_choppedMaize_sml.png"

    local silagePrice = fillTypeManager:getFillTypeByName("SILAGE").pricePerLiter
    local chaffMass = fillTypeManager:getFillTypeByName("CHAFF").massPerLiter

    local fillType = fillTypeManager:addFillType("CHOPPEDMAIZE", i18n:getText("fillType_choppedMaize"), true, silagePrice * 0.3, chaffMass, 35, hudFile, hudFileSmall, directory, nil, { 1, 1, 1 }, nil, false)
	
    if fillType == nil then
		print("MaizePlus: Error, could not add FillType choppedMaize.")
    end
end

function MaizePlus.loadChoppedMaizeFermentedFillType(directory, fillTypeManager, i18n)
    local hudFile = "hud/hud_fill_choppedMaize_fermented.png"
    local hudFileSmall = "hud/hud_fill_choppedMaize_fermented_sml.png"

    local silagePrice = fillTypeManager:getFillTypeByName("SILAGE").pricePerLiter
    local silageMass = fillTypeManager:getFillTypeByName("SILAGE").massPerLiter

    local fillType = fillTypeManager:addFillType("CHOPPEDMAIZE_FERMENTED", i18n:getText("fillType_choppedMaize_fermented"), true, silagePrice, silageMass, 50, hudFile, hudFileSmall, directory, nil, { 1, 1, 1 }, nil, false)
	
    if fillType == nil then
		print("MaizePlus: Error, could not add FillType choppedMaize_fermented.")
    end
end

function MaizePlus.loadGrassFermentedFillType(directory, fillTypeManager, i18n)
    local hudFile = "hud/hud_fill_grass_fermented.png"
    local hudFileSmall = "hud/hud_fill_grass_fermented_sml.png"

    local silagePrice = fillTypeManager:getFillTypeByName("SILAGE").pricePerLiter
    local silageMass = fillTypeManager:getFillTypeByName("SILAGE").massPerLiter

    local fillType = fillTypeManager:addFillType("GRASS_FERMENTED", i18n:getText("fillType_grass_fermented"), true, silagePrice*0.8, silageMass*0.75, 50, hudFile, hudFileSmall, directory, nil, { 1, 1, 1 }, nil, false)
	
    if fillType == nil then
		print("MaizePlus: Error, could not add FillType grass_fermented.")
    end
end

function MaizePlus.loadSiloGrass(directory, fillTypeManager, i18n)
    local hudFile = "hud/hud_fill_siloGrass.png"
    local hudFileSmall = "hud/hud_fill_siloGrass_sml.png"

    local price = fillTypeManager:getFillTypeByName("GRASS_WINDROW").pricePerLiter
	local massPerLiter = fillTypeManager:getFillTypeByName("GRASS_WINDROW").massPerLiter

    local fillType = fillTypeManager:addFillType("SILOGRASS", i18n:getText("fillType_siloGrass"), true, price, massPerLiter, 35, hudFile, hudFileSmall, directory, nil, { 1, 1, 1 }, nil, false)
	
    if fillType == nil then
		print("MaizePlus: Error, could not add FillType CCMRAW.")
    end
end

function MaizePlus.loadCCMFillType(directory, fillTypeManager, i18n)
    local hudFile = "hud/hud_fill_ccm.png"
    local hudFileSmall = "hud/hud_fill_ccm_sml.png"

    local maizePrice = fillTypeManager:getFillTypeByName("MAIZE").pricePerLiter
	local massPerLiter = 0.00043

    local fillType = fillTypeManager:addFillType("CCM", i18n:getText("fillType_ccm"), true, maizePrice*2, massPerLiter, 40, hudFile, hudFileSmall, directory, nil, { 1, 1, 1 }, nil, false)
	
    if fillType == nil then
		print("MaizePlus: Error, could not add FillType CCM.")
    end
end

function MaizePlus.loadCCMRawFillType(directory, fillTypeManager, i18n)
    local hudFile = "hud/hud_fill_ccmraw.png"
    local hudFileSmall = "hud/hud_fill_ccmraw_sml.png"

    local maizePrice = fillTypeManager:getFillTypeByName("MAIZE").pricePerLiter
	local massPerLiter = 0.0004

    local fillType = fillTypeManager:addFillType("CCMRAW", i18n:getText("fillType_ccmraw"), true, maizePrice*1.2, massPerLiter, 30, hudFile, hudFileSmall, directory, nil, { 1, 1, 1 }, nil, false)
	
    if fillType == nil then
		print("MaizePlus: Error, could not add FillType CCMRAW.")
    end
end



---Add chopped maize fillType.
function MaizePlus.addChoppedMaize()

    local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE")
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["WINDROW"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["BULK"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEWAGON"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEHARVESTER"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["SILAGETRAILER"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORK"])
	
	g_fillTypeManager.fillTypeToSample[fillTypeIndex] = g_fillTypeManager.fillTypeToSample[g_fillTypeManager:getFillTypeIndexByName("CHAFF")]	

    local diffuseMapFilename = Utils.getFilename("fillTypes/choppedMaize_diffuse.png", modDirectory)
    local normalMapFilename = Utils.getFilename("fillTypes/chaff_normal.png", modDirectory)
    local distanceMapFilename = Utils.getFilename("fillTypes/choppedMaize_distance_diffuse.png", modDirectory)

    local choppedMaize_heightType = g_densityMapHeightManager:addDensityMapHeightType("CHOPPEDMAIZE", math.rad(35), 0.8, 0.1, 0.1, 1.2, 1.1, true, diffuseMapFilename, normalMapFilename, distanceMapFilename, false)
		
	if choppedMaize_heightType ~= nil then
        return true
    end

    return false
end

---Add chopped maize fermented fillType.
function MaizePlus.addChoppedMaizeFermented()
    local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE_FERMENTED")
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["WINDROW"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["BULK"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEWAGON"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEHARVESTER"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["SILAGETRAILER"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORK"])
	
	g_fillTypeManager.fillTypeToSample[fillTypeIndex] = g_fillTypeManager.fillTypeToSample[g_fillTypeManager:getFillTypeIndexByName("SILAGE")]		

    local diffuseMapFilename2 = Utils.getFilename("fillTypes/choppedMaize_fermented_diffuse.png", modDirectory)
    local normalMapFilename2 = Utils.getFilename("fillTypes/silage_normal.png", modDirectory)
    local distanceMapFilename2 = Utils.getFilename("fillTypes/silage_distance_diffuse.png", modDirectory)

    local choppedMaize_fermented_heightType = g_densityMapHeightManager:addDensityMapHeightType("CHOPPEDMAIZE_FERMENTED", math.rad(50), 0.95, 0.0, 0.0, 0.0, 1.0, true, diffuseMapFilename2, normalMapFilename2, distanceMapFilename2, false)
    if choppedMaize_fermented_heightType ~= nil then
        return true
    end

    return false
end

---Add chopped maize fermented fillType.
function MaizePlus.addGrassFermented()
    local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName("GRASS_FERMENTED")
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["WINDROW"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["BULK"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEWAGON"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEHARVESTER"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["SILAGETRAILER"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORK"])
	
	g_fillTypeManager.fillTypeToSample[fillTypeIndex] = g_fillTypeManager.fillTypeToSample[g_fillTypeManager:getFillTypeIndexByName("SILAGE")]		

    local diffuseMapFilename2 = Utils.getFilename("fillTypes/grass_fermented_diffuse.png", modDirectory)
    local normalMapFilename2 = Utils.getFilename("fillTypes/grass_fermented_normal.png", modDirectory)
    local distanceMapFilename2 = Utils.getFilename("fillTypes/grass_fermented_distance_diffuse.png", modDirectory)

    local grass_fermented_heightType = g_densityMapHeightManager:addDensityMapHeightType("GRASS_FERMENTED", math.rad(50), 0.95, 0.0, 0.0, 0.0, 1.0, true, diffuseMapFilename2, normalMapFilename2, distanceMapFilename2, false)
    if grass_fermented_heightType ~= nil then
        return true
    end

    return false
end

---Add chopped maize fermented fillType.
function MaizePlus.addCCM()
	if not MaizePlus.ccmIsEnabled then
		return true
	end
    local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName("CCM")
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["WINDROW"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["BULK"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEWAGON"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEHARVESTER"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["SILAGETRAILER"])	
	
	g_fillTypeManager.fillTypeToSample[fillTypeIndex] = g_fillTypeManager.fillTypeToSample[g_fillTypeManager:getFillTypeIndexByName("SILAGE")]		

    local diffuseMapFilename2 = Utils.getFilename("fillTypes/ccm_diffuse.png", modDirectory)
    local normalMapFilename2 = Utils.getFilename("fillTypes/ccm_normal.png", modDirectory)
    local distanceMapFilename2 = Utils.getFilename("fillTypes/ccm_distance_diffuse.png", modDirectory)

    local ccm_heightType = g_densityMapHeightManager:addDensityMapHeightType("CCM", math.rad(40), 0.95, 0.0, 0.0, 0.0, 1.0, true, diffuseMapFilename2, normalMapFilename2, distanceMapFilename2, false)
    if ccm_heightType ~= nil then
        return true
    end

    return false
end

---Add chopped maize fermented fillType.
function MaizePlus.addCCMRaw()
	if not MaizePlus.ccmIsEnabled then
		return true
	end
    local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName("CCMRAW")
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["WINDROW"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["BULK"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEWAGON"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEHARVESTER"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["SILAGETRAILER"])	
	
	g_fillTypeManager.fillTypeToSample[fillTypeIndex] = g_fillTypeManager.fillTypeToSample[g_fillTypeManager:getFillTypeIndexByName("CHAFF")]		

    local diffuseMapFilename2 = Utils.getFilename("fillTypes/ccmraw_diffuse.png", modDirectory)
    local normalMapFilename2 = Utils.getFilename("fillTypes/ccm_normal.png", modDirectory)
    local distanceMapFilename2 = Utils.getFilename("fillTypes/ccmraw_distance_diffuse.png", modDirectory)

    local ccmraw_heightType = g_densityMapHeightManager:addDensityMapHeightType("CCMRAW", math.rad(30), 0.8, 0.1, 0.1, 1.2, 1.0, true, diffuseMapFilename2, normalMapFilename2, distanceMapFilename2, false)
    if ccmraw_heightType ~= nil then
        return true
    end

    return false
end

---Add chopped maize fermented fillType.
function MaizePlus.addSiloGrass()
    local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName("SILOGRASS")
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["WINDROW"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["BULK"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEWAGON"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORAGEHARVESTER"])
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["SILAGETRAILER"])	
    g_fillTypeManager:addFillTypeToCategory(fillTypeIndex, g_fillTypeManager.nameToCategoryIndex["FORK"])	
	
	g_fillTypeManager.fillTypeToSample[fillTypeIndex] = g_fillTypeManager.fillTypeToSample[g_fillTypeManager:getFillTypeIndexByName("GRASS_WINDROW")]		

    local diffuseMapFilename2 = Utils.getFilename("fillTypes/siloGrass_diffuse.png", modDirectory)
    local normalMapFilename2 = Utils.getFilename("fillTypes/grass_fermented_normal.png", modDirectory)
    local distanceMapFilename2 = Utils.getFilename("fillTypes/siloGrass_distance_diffuse.png", modDirectory)
	

    local ccmraw_heightType = g_densityMapHeightManager:addDensityMapHeightType("SILOGRASS", math.rad(35), 0.8, 0.1, 0.1, 1.2, 1.3, true, diffuseMapFilename2, normalMapFilename2, distanceMapFilename2, false)
    if ccmraw_heightType ~= nil then
        return true
    end

    return false
end			

---Hook into the loadMapFinished to load the fillTypes, material holders and add new bale types.
function MaizePlus.initFillTypes(mission, node)

    if MaizePlus.addChoppedMaize() and MaizePlus.addChoppedMaizeFermented() and MaizePlus.addGrassFermented() and MaizePlus.addCCM() and MaizePlus.addCCMRaw() and MaizePlus.addSiloGrass() then
        print("MaizePlus: Filltypes Added.")
    else
        print("MaizePlus: Error, something went wrong with adding additional height types.")
		print("You most likely have too many high types already registered.")
		print("MaizePlus won't work properly.")
		MaizePlus.isActive = false
		return
    end

    -- refreshing the vehicle fillLevels HUD
    g_currentMission.hud.fillLevelsDisplay:refreshFillTypes(g_fillTypeManager)

    -- add chopped maize to forage harvester converter
    g_fruitTypeManager.nameToConverter.FORAGEHARVESTER[g_fruitTypeManager.nameToIndex.MAIZE].fillTypeIndex = g_fillTypeManager.nameToIndex.CHOPPEDMAIZE

	-- change grass to not be chaff out of a chopper 
    g_fruitTypeManager.nameToConverter.FORAGEHARVESTER[g_fruitTypeManager.nameToIndex.GRASS].fillTypeIndex = g_fillTypeManager.nameToIndex.GRASS_WINDROW
	
	-- change names of default silage and chaff 
	g_fillTypeManager.fillTypes[g_fillTypeManager.nameToIndex.SILAGE].title = g_i18n:getText("fillType_silageNew")
	g_fillTypeManager.fillTypes[g_fillTypeManager.nameToIndex.CHAFF].title = g_i18n:getText("fillType_chaffNew")
    

    -- loading material holders 
    local maizePlus_materialHolderFilename = Utils.getFilename("materialHolder/maizePlus_materialHolder.i3d", modDirectory)
    MaizePlus.maizePlus_materialHolder = loadI3DFile(maizePlus_materialHolderFilename, false, true, false)
	
    local maizePlus_effect_materialHolderFilename = Utils.getFilename("materialHolder/maizePlus_effect_materialHolder.i3d", modDirectory)
    MaizePlus.maizePlus_effect_materialHolder = loadI3DFile(maizePlus_effect_materialHolderFilename, false, true, false)
	
	if MaizePlus.ccmIsEnabled then
	    local maizePlus_ccmExtension_materialHolderFilename = Utils.getFilename("materialHolder/maizePlus_ccmExtension_materialHolder.i3d", modDirectory)
	    MaizePlus.maizePlus_ccmExtension_materialHolder = loadI3DFile(maizePlus_ccmExtension_materialHolderFilename, false, true, false)
	    local maizePlus_ccmExtension_effect_materialHolderFilename = Utils.getFilename("materialHolder/maizePlus_ccmExtension_effect_materialHolder.i3d", modDirectory)
	    MaizePlus.maizePlus_ccmExtension_effect_materialHolder = loadI3DFile(maizePlus_ccmExtension_effect_materialHolderFilename, false, true, false)		
	end

    -- add chopped maize bale type
    g_baleTypeManager:addBaleType(Utils.getFilename("roundbales/roundbale_choppedMaize_w112_d130.i3d", modDirectory), "CHOPPEDMAIZE", true, 1.12, 0.9, 2.4, 1.3)
    g_baleTypeManager:addBaleType(Utils.getFilename("roundbales/roundbale_silograss_w112_d130.i3d", modDirectory), "SILOGRASS", true, 1.12, 0.9, 2.4, 1.3)

    -- add chopped maize fermented bale types
    g_baleTypeManager:addBaleType(Utils.getFilename("roundbales/roundbale_choppedMaize_fermented_w112_d130.i3d", modDirectory), "CHOPPEDMAIZE_FERMENTED", true, 1.12, 0.9, 2.4, 1.3)
    g_baleTypeManager:addBaleType(Utils.getFilename("roundbales/kuhnFBP3135_roundbale_choppedMaize_fermented_w112_d130.i3d", modDirectory), "CHOPPEDMAIZE_FERMENTED", true, 1.12, 0.9, 2.4, 1.3)
    g_baleTypeManager:addBaleType(Utils.getFilename("roundbales/kuhnSW4014_roundbale_choppedMaize_fermented_w112_d130.i3d", modDirectory), "CHOPPEDMAIZE_FERMENTED", true, 1.12, 0.9, 2.4, 1.3)

	-- add grass fermented bale types 
	g_baleTypeManager:addBaleType(Utils.getFilename("roundbales/roundbale_grass_fermented_w112_d130.i3d", modDirectory), "GRASS_FERMENTED", true, 1.12, 0.9, 2.4, 1.3)
	g_baleTypeManager:addBaleType(Utils.getFilename("roundbales/kuhnFBP3135_roundbale_grass_fermented_w112_d130.i3d", modDirectory), "GRASS_FERMENTED", true, 1.12, 0.9, 2.4, 1.3)
	g_baleTypeManager:addBaleType(Utils.getFilename("roundbales/kuhnSW4014_roundbale_grass_fermented_w112_d130.i3d", modDirectory), "GRASS_FERMENTED", true, 1.12, 0.9, 2.4, 1.3)

	-- add square grass bale for fillType silograss
	g_baleTypeManager:addBaleType(Utils.getFilename("squarebales/squarebale_siloGrass.i3d", modDirectory), "SILOGRASS", false, 1.20, 0.9, 2.4)


end


---Change baler so they can bale non-windrow choppedMaize
function MaizePlus.balerOnPostLoadAdd(vehicle, savegame)
	if MaizePlus.isActive then
		local spec = vehicle.spec_baler
		for fillTypeIndex, enabled in pairs(vehicle:getFillUnitSupportedFillTypes(spec.fillUnitIndex)) do
			if enabled and fillTypeIndex ~= FillType.UNKNOWN then
				if fillTypeIndex == FillType.CHOPPEDMAIZE then
					table.insert(spec.pickupFillTypes, fillTypeIndex)
				end
				if fillTypeIndex == FillType.SILOGRASS then
					table.insert(spec.pickupFillTypes, fillTypeIndex)
				end				
			end
		end
	end
end

---Add choppedMaize_fermented to "silage" category on mixer wagon
function MaizePlus.mixerWagonOnLoadAdd(vehicle, savegame)
	if MaizePlus.isActive then
		local spec = vehicle.spec_mixerWagon

		local fillTypeIndex_choppedMaize = g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE")
		local fillTypeIndex_choppedMaize_fermented = g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE_FERMENTED")
		local fillTypeIndex_grass_fermented = g_fillTypeManager:getFillTypeIndexByName("GRASS_FERMENTED")

		-- add choppedMaize and choppedMaize_fermented and grass_fermented and grass_fermented to mixer wagon fillTypes
		for fillUnitIndex, fillUnit in pairs(vehicle.spec_fillUnit.fillUnits) do
			-- go through fillUnits
			if fillUnit.supportedFillTypes[g_fillTypeManager:getFillTypeIndexByName("SILAGE")] then
				-- if we have a fillUnit that supports silage, we found the right one (99%)
				fillUnit.supportedFillTypes[fillTypeIndex_choppedMaize] = true
				fillUnit.supportedFillTypes[fillTypeIndex_choppedMaize_fermented] = true
				fillUnit.supportedFillTypes[fillTypeIndex_grass_fermented] = true
			end
		end

		-- add choppedMaize_fermented to "silage" category in mixerWagon, add choppedMaize to "dryGrass" category in mixer wagon
		for _, fillType in pairs(spec.mixerWagonFillTypes) do
			if fillType.name == "silage" then
				if spec.fillTypeToMixerWagonFillType[fillTypeIndex_choppedMaize_fermented] == nil then
					fillType.fillTypes[fillTypeIndex_choppedMaize_fermented] = true
					spec.fillTypeToMixerWagonFillType[fillTypeIndex_choppedMaize_fermented] = fillType
				end
				if spec.fillTypeToMixerWagonFillType[fillTypeIndex_grass_fermented] == nil then
					fillType.fillTypes[fillTypeIndex_grass_fermented] = true
					spec.fillTypeToMixerWagonFillType[fillTypeIndex_grass_fermented] = fillType
				end				
			end
			if fillType.name == "dryGrass" then
				if spec.fillTypeToMixerWagonFillType[fillTypeIndex_choppedMaize] == nil then
					fillType.fillTypes[fillTypeIndex_choppedMaize] = true
					spec.fillTypeToMixerWagonFillType[fillTypeIndex_choppedMaize] = fillType
				end
			end
		end
	end
end

---Change bale wrapper so they will wrap choppedmaize bales into choppedmaize_fermented bales
function MaizePlus.loadWrapperFromXMLAdd(self, wrapper, xmlFile, baseKey)
	if MaizePlus.isActive then
		-- only add maize bale to roundbale wrapper 
		if wrapper.allowedBaleTypes ~= nil and wrapper.allowedBaleTypes[FillType.GRASS_WINDROW] ~= nil then
			
			-- round bale wrapper 
			if wrapper.allowedBaleTypes[FillType.GRASS_WINDROW][1].minBaleDiameter ~= nil then
			
				-- we need to check if we are the default KuhnFBP3135 or KuhnSW4014 because those have special bales 
				-- (we would need to do the same for the Claas baler but we can't since that bale is locked away)
				-- split path and check for the name in the path 
				local isKuhnSW4014 = false
				local isKuhnFBP3135 = false
				local splitPath = StringUtil.splitString("/", wrapper.allowedBaleTypes[FillType.GRASS_WINDROW][1].wrapperBaleFilename)
				for _, pathPart in pairs(splitPath) do
					if pathPart == "kuhnSW4014" then 
						isKuhnSW4014 = true
					end
					if pathPart == "kuhnFBP3135" then
						isKuhnFBP3135 = true
					end
				end
		
				-- set up path depending on which wrapper we are 
				local roundbale_choppedMaize_fermented_w112_d130 = Utils.getFilename("roundbales/roundbale_choppedMaize_fermented_w112_d130.i3d", modDirectory)
				local roundbale_grass_fermented_w112_d130 = Utils.getFilename("roundbales/roundbale_grass_fermented_w112_d130.i3d", modDirectory)
				if isKuhnFBP3135 then
					roundbale_choppedMaize_fermented_w112_d130 = Utils.getFilename("roundbales/kuhnFBP3135_roundbale_choppedMaize_fermented_w112_d130.i3d", modDirectory)
					roundbale_grass_fermented_w112_d130 = Utils.getFilename("roundbales/kuhnFBP3135_roundbale_grass_fermented_w112_d130.i3d", modDirectory)
				elseif isKuhnSW4014 then
					roundbale_choppedMaize_fermented_w112_d130 = Utils.getFilename("roundbales/kuhnSW4014_roundbale_choppedMaize_fermented_w112_d130.i3d", modDirectory)
					roundbale_grass_fermented_w112_d130 = Utils.getFilename("roundbales/kuhnSW4014_roundbale_grass_fermented_w112_d130.i3d", modDirectory)
				end
				-- add chopped maize to allowed bale types and add bale 
				wrapper.allowedBaleTypes[FillType.CHOPPEDMAIZE] = {}
				table.insert(wrapper.allowedBaleTypes[FillType.CHOPPEDMAIZE], { fillType = FillType.CHOPPEDMAIZE, wrapperBaleFilename = roundbale_choppedMaize_fermented_w112_d130, minBaleDiameter = 1.1, maxBaleDiameter = 1.3, minBaleWidth = 1.1, maxBaleWidth = 1.3 })
				
				-- add grass silage bale to grass 
				wrapper.allowedBaleTypes[FillType.GRASS_WINDROW][1].wrapperBaleFilename = roundbale_grass_fermented_w112_d130

				-- add silograss to allowed bale types and add bale 
				wrapper.allowedBaleTypes[FillType.SILOGRASS] = {}
				table.insert(wrapper.allowedBaleTypes[FillType.SILOGRASS], { fillType = FillType.SILOGRASS, wrapperBaleFilename = roundbale_grass_fermented_w112_d130, minBaleDiameter = 1.1, maxBaleDiameter = 1.3, minBaleWidth = 1.1, maxBaleWidth = 1.3 })
								
			elseif wrapper.allowedBaleTypes[FillType.GRASS_WINDROW][1].minBaleLength ~= nil then -- square bale wrapper
				print("Is square bale wrapper")
				-- square silage bale from Kuhn bale wrapper, use this for all square balers if there's any
				local squarebale_grass_fermented = Utils.getFilename("squarebales/squarebale_grass_fermented.i3d", modDirectory)

				-- add silograss to allowed bale types and add bale 
				wrapper.allowedBaleTypes[FillType.SILOGRASS] = {}
				table.insert(wrapper.allowedBaleTypes[FillType.SILOGRASS], { fillType = FillType.SILOGRASS, wrapperBaleFilename = squarebale_grass_fermented, minBaleWidth = 1.15, maxBaleWidth = 1.25, maxBaleHeight = 0.85, minBaleHeight = 0.75, minBaleLength = 2.35, maxBaleLength = 2.45})
				
				-- add grass silage bale to grass 
				wrapper.allowedBaleTypes[FillType.GRASS_WINDROW][1].wrapperBaleFilename = squarebale_grass_fermented									
			end
		end
	end
end

---Add choppedmaize and choppedmaize_fermented to animal food groups so it can be fed
function MaizePlus.loadFoodGroupsAdd()
	if MaizePlus.isActive then
		local foodManager = g_animalFoodManager
		
		-- go through all food groups for "cow"
		local animalType = "COW"
		for _, foodGroup in pairs(foodManager.foodGroups[animalType].content) do
			for _, fillType in pairs(foodGroup.fillTypes) do
				-- go through all the fillTypes for the current food group
				if fillType == g_fillTypeManager:getFillTypeIndexByName("SILAGE") then
					-- check if we have silage listed
					-- add choppedMaize_fermented to silage category
					table.insert(foodGroup.fillTypes, g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE_FERMENTED"))
				end
				if fillType == g_fillTypeManager:getFillTypeIndexByName("GRASS_WINDROW") then
					-- check if we have grass listed
					-- add choppedMaize grass category
					table.insert(foodGroup.fillTypes, g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE"))
					table.insert(foodGroup.fillTypes, g_fillTypeManager:getFillTypeIndexByName("GRASS_FERMENTED"))
				end
			end
		end

		-- same for sheep
		animalType = "SHEEP"
		for _, foodGroup in pairs(foodManager.foodGroups[animalType].content) do
			for _, fillType in pairs(foodGroup.fillTypes) do
				-- go through all the fillTypes for the current food group
				if fillType == g_fillTypeManager:getFillTypeIndexByName("GRASS_WINDROW") then
					-- check if we have grass listed
					-- add choppedMaize and choppedMaize_fermented grass category
					table.insert(foodGroup.fillTypes, g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE"))
					table.insert(foodGroup.fillTypes, g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE_FERMENTED"))
					table.insert(foodGroup.fillTypes, g_fillTypeManager:getFillTypeIndexByName("GRASS_FERMENTED"))
				end
			end
		end

		-- pigs.. now pigs are a different kind of animal )
		animalType = "PIG"
		for _, foodGroup in pairs(foodManager.foodGroups[animalType].content) do
			for _, fillType in pairs(foodGroup.fillTypes) do
				-- go through all the fillTypes for the current food group
				if fillType == g_fillTypeManager:getFillTypeIndexByName("WHEAT") then
					-- check if we have wheat listed
					-- add choppedMaize to wheat category
					table.insert(foodGroup.fillTypes, g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE"))
				end
				if MaizePlus.ccmIsEnabled and fillType == g_fillTypeManager:getFillTypeIndexByName("MAIZE") then
					-- check if we have maize listed
					-- add ccm to maize category
					table.insert(foodGroup.fillTypes, g_fillTypeManager:getFillTypeIndexByName("CCM"))
				end				
			end
		end
		
		animalType = "CHICKEN"
		for _, foodGroup in pairs(foodManager.foodGroups[animalType].content) do
			for _, fillType in pairs(foodGroup.fillTypes) do
				-- go through all the fillTypes for the current food group
				if MaizePlus.ccmIsEnabled and fillType == g_fillTypeManager:getFillTypeIndexByName("WHEAT") then
					-- check if we have maize listed
					-- add ccm to maize category
					table.insert(foodGroup.fillTypes, g_fillTypeManager:getFillTypeIndexByName("CCMRAW"))
				end				
			end
		end		
		
	end
end

---Add choppedMaize to selling stations that sell grass and add choppedMaize_fermented to selling stations that sell silage
function MaizePlus.sellingStationLoad(self, xmlFile, key, customEnvironment)
	if MaizePlus.isActive then
		if self ~= nil then
			local choppedMaizeIndex = g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE")
			local choppedMaizeFermentedIndex = g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE_FERMENTED")
			local ccmrawIndex = g_fillTypeManager:getFillTypeIndexByName("CCMRAW")
			local ccmIndex = g_fillTypeManager:getFillTypeIndexByName("CCM")
			local grassFermentedIndex = g_fillTypeManager:getFillTypeIndexByName("GRASS_FERMENTED")

			-- check if selling station sells grass
			if self.acceptedFillTypes[g_fillTypeManager:getFillTypeIndexByName("GRASS_WINDROW")] then
				-- add chopped maize fill type
				self:addAcceptedFillType(choppedMaizeIndex, g_fillTypeManager:getFillTypeByName("CHOPPEDMAIZE").pricePerLiter, true, false)
				self.priceMultipliers[choppedMaizeIndex] = self.priceMultipliers[g_fillTypeManager:getFillTypeIndexByName("GRASS_WINDROW")]
				
				self:addAcceptedFillType(ccmrawIndex, g_fillTypeManager:getFillTypeByName("CCMRAW").pricePerLiter, true, false)
				self.priceMultipliers[ccmrawIndex] = self.priceMultipliers[g_fillTypeManager:getFillTypeIndexByName("GRASS_WINDROW")]
				
				self:initPricingDynamics()
			end

			-- check if selling station sells silage
			if self.acceptedFillTypes[g_fillTypeManager:getFillTypeIndexByName("SILAGE")] then
				-- add chopped maize fermented  fill type

				self:addAcceptedFillType(choppedMaizeFermentedIndex, g_fillTypeManager:getFillTypeByName("CHOPPEDMAIZE_FERMENTED").pricePerLiter, true, false)
				self.priceMultipliers[choppedMaizeFermentedIndex] = self.priceMultipliers[g_fillTypeManager:getFillTypeIndexByName("SILAGE")]
				
				self:addAcceptedFillType(grassFermentedIndex, g_fillTypeManager:getFillTypeByName("GRASS_FERMENTED").pricePerLiter, true, false)
				self.priceMultipliers[grassFermentedIndex] = self.priceMultipliers[g_fillTypeManager:getFillTypeIndexByName("SILAGE")]	

				self:addAcceptedFillType(ccmIndex, g_fillTypeManager:getFillTypeByName("CCM").pricePerLiter, true, false)
				self.priceMultipliers[ccmIndex] = self.priceMultipliers[g_fillTypeManager:getFillTypeIndexByName("SILAGE")]		
				
				self:initPricingDynamics()
			end
		end
	end
end

-- add choppedMaize_fermented to BGA bunker stations
function MaizePlus.bgaLoad(self, superFunc, xmlFile, key, customEnvironment)
    local returnValue = superFunc(self, xmlFile, key, customEnvironment)

	if MaizePlus.isActive then
		if self ~= nil then

			local silageIndex = g_fillTypeManager:getFillTypeIndexByName("SILAGE")
			local choppedMaizeFermentedIndex = g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE_FERMENTED")
			local choppedMaizeIndex = g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE")
			local grassWindrowIndex = g_fillTypeManager:getFillTypeIndexByName("GRASS_WINDROW")
			local grassFermentedIndex = g_fillTypeManager:getFillTypeIndexByName("GRASS_FERMENTED")
		
			-- go through all BGA slots and check their fillTypes for "silage"
			for _, slot in pairs(self.bunker.slots) do

				if slot.fillTypes[silageIndex] then
					-- add attributes for new fillType
					-- add choppedMaize_fermented
					local fillType = {}
					fillType.fillLevel = 0
					fillType.scale = slot.fillTypes[silageIndex].scale
					fillType.pricePerLiter = slot.fillTypes[silageIndex].pricePerLiter

					-- add new fillType to fillTypes
					slot.fillTypes[choppedMaizeFermentedIndex] = fillType

					-- add slot to fillType-to-slot table
					if self.bunker.fillTypeToSlot[choppedMaizeFermentedIndex] == nil then
						self.bunker.fillTypeToSlot[choppedMaizeFermentedIndex] = slot
					end
					
					fillType = {}
					-- add grass_fermented
					fillType.fillLevel = 0
					fillType.scale = slot.fillTypes[silageIndex].scale * 0.8
					fillType.pricePerLiter = slot.fillTypes[silageIndex].pricePerLiter * 0.8

					-- add new fillType to fillTypes
					slot.fillTypes[grassFermentedIndex] = fillType

					-- add slot to fillType-to-slot table
					if self.bunker.fillTypeToSlot[grassFermentedIndex] == nil then
						self.bunker.fillTypeToSlot[grassFermentedIndex] = slot
					end				
				end

				if slot.fillTypes[grassWindrowIndex] then
					-- add attributes for new fillType
					-- add raw choppedMaize
					local fillType = {}
					fillType.fillLevel = 0
					fillType.scale = slot.fillTypes[grassWindrowIndex].scale
					fillType.pricePerLiter = slot.fillTypes[grassWindrowIndex].pricePerLiter

					-- add new fillType to fillTypes
					slot.fillTypes[g_fillTypeManager:getFillTypeIndexByName("CHOPPEDMAIZE")] = fillType

					-- add slot to fillType-to-slot table
					if self.bunker.fillTypeToSlot[choppedMaizeIndex] == nil then
						self.bunker.fillTypeToSlot[choppedMaizeIndex] = slot
					end
				end

			end

		end
	end
	
    return returnValue
end

-- load fillTypes
local function registerFillTypes()
    MaizePlus.fillTypeInit(modDirectory, g_fillTypeManager, g_i18n)
end


---Unload the mod when the game is closed.
local function unload()
	if MaizePlus.maizePlus_ccmExtension_materialHolder ~= nil then
		delete(MaizePlus.maizePlus_ccmExtension_materialHolder)
		MaizePlus.maizePlus_ccmExtension_materialHolder = nil
	end
	if MaizePlus.maizePlus_ccmExtension_effect_materialHolder ~= nil then
		delete(MaizePlus.maizePlus_ccmExtension_effect_materialHolder)
		MaizePlus.maizePlus_ccmExtension_effect_materialHolder = nil
	end	
	if MaizePlus.maizePlus_effect_materialHolder ~= nil then
		delete(MaizePlus.maizePlus_effect_materialHolder)
		MaizePlus.maizePlus_effect_materialHolder = nil
	end	
	if MaizePlus.maizePlus_materialHolder ~= nil then
		delete(MaizePlus.maizePlus_materialHolder)
		MaizePlus.maizePlus_materialHolder = nil
	end	
end

---Called on first time the MaizePlus script is loaded.
local function init()
    -- Init the fill types that will be used on the heightmap earlier.
    FillTypeManager.loadMapData = Utils.appendedFunction(FillTypeManager.loadMapData, registerFillTypes)
	
    -- Add fillTypes before loadMapFinished to not get stuff mixed up
    FSBaseMission.loadMap = Utils.prependedFunction(FSBaseMission.loadMap, MaizePlus.initFillTypes)

    -- unload/delete stuff before game closes 
    FSBaseMission.delete = Utils.appendedFunction(FSBaseMission.delete, unload)

	-- hook into loadMapFinished to edit food groups since when foodManager is loaded our fillTypes aren't loaded yet 
    FSBaseMission.loadMapFinished = Utils.appendedFunction(FSBaseMission.loadMapFinished, MaizePlus.loadFoodGroupsAdd)
	
	-- add our stuff to selling stations 
    SellingStation.load = Utils.appendedFunction(SellingStation.load, MaizePlus.sellingStationLoad)

	-- add our stuff to BGA 
    Bga.load = Utils.overwrittenFunction(Bga.load, MaizePlus.bgaLoad)
	
	-- a lot of changes to bunkerSilo, thats why its in a separate file now 
    BunkerSilo.load = Utils.overwrittenFunction(BunkerSilo.load, MaizePlus_bunkerSiloChanges.bunkerSiloLoad)
    BunkerSilo.interactionTriggerCallback = Utils.appendedFunction(BunkerSilo.interactionTriggerCallback, MaizePlus_bunkerSiloChanges.bunkerSiloInteractionTriggerCallback)
    BunkerSilo.saveToXMLFile = Utils.appendedFunction(BunkerSilo.saveToXMLFile, MaizePlus_bunkerSiloChanges.bunkerSiloSaveToXMLFile)
    BunkerSilo.loadFromXMLFile = Utils.overwrittenFunction(BunkerSilo.loadFromXMLFile, MaizePlus_bunkerSiloChanges.bunkerSiloLoadFromXMLFile)
	BunkerSilo.updateTick = Utils.appendedFunction(BunkerSilo.updateTick, MaizePlus_bunkerSiloChanges.updateTick)
	
	BunkerSilo.writeStream = Utils.appendedFunction(BunkerSilo.writeStream, MaizePlus_bunkerSiloChanges.writeStream)	
	BunkerSilo.readStream = Utils.appendedFunction(BunkerSilo.readStream, MaizePlus_bunkerSiloChanges.readStream)	

	-- add stuff to vehicles. (no extra specs since I don't think its neccessary in our case, if somebody needs to remove/change something in order for his mod to work, just talk to us)
    Baler.onPostLoad = Utils.appendedFunction(Baler.onPostLoad, MaizePlus.balerOnPostLoadAdd)
    MixerWagon.onLoad = Utils.appendedFunction(MixerWagon.onLoad, MaizePlus.mixerWagonOnLoadAdd)
    BaleWrapper.loadWrapperFromXML = Utils.appendedFunction(BaleWrapper.loadWrapperFromXML, MaizePlus.loadWrapperFromXMLAdd)
end

init()


-- Thank you Wopster, again ;) 